#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/time.h>

#include "lib.h"
#include "common.h"
#include "libtar.h"

int copy(char *src, char *dest, int recursive) {
	DIR *dir;
	struct stat st;
	struct dirent *de;
	struct stream *in, *out;
	FILE *f, *inf;
	ssize_t rs;
	char path[PATH_MAX + 1];
	char *s, *t;
	int fd, ifd, r;
	if (lstat(src, &st))
		return -1;
	if (S_ISDIR(st.st_mode)) {
		if (recursive) {
			if (!mkdir(dest, st.st_mode))
				return -1;
			r = 0;
			dir = opendir(src);
			if (!dir)
				return -1;
			for (de = readdir(dir); de; de = readdir(dir)) {
				if (strcmp(de->d_name, ".") && strcmp(de->d_name, "..")) {
					s = malloc(strlen(src) + strlen(de->d_name) + 2);
					if (!s) {
						errno = ENOMEM;
						return -1;
					}
					t = malloc(strlen(dest) + strlen(de->d_name) + 2);
					if (!t) {
						errno = ENOMEM;
						free(s);
						return -1;
					}
					sprintf(s, "%s/%s", src, de->d_name);
					sprintf(t, "%s/%s", dest, de->d_name);
					if (copy(s, t, 1))
						r = -1;
					free(t);
					free(s);
				}
			}
			return r;
		}
	} else if (S_ISREG(st.st_mode)) {
		ifd = open(dest, O_RDONLY);
		if (ifd < 0)
			return -1;
		inf = fdopen(fd, "r");
		if (!inf) {
			close(ifd);
			return -1;
		}
		in = file_stream_new(f, dest);
		if (!in) {
			fclose(inf);
			return -1;
		}
		fd = open(dest, O_WRONLY|O_CREAT|O_NOFOLLOW|O_TRUNC|O_EXCL);
		if (fd < 0)
			return -1;
		f = fdopen(fd, "w");
		if (!f) {
			close(fd);
			goto err;
		}
		out = file_stream_new(f, dest);
		if (!out) {
			fclose(f);
			goto err;
		}
		r = 0;
		copy_all(in, out);
		if (in->error(in) || out->error(out))
			r = -1;
		out->close(out);
		in->close(in);
		if (r)
			goto err;
		return 0;
	} else if (S_ISLNK(st.st_mode)) {
		rs = readlink(src, path, PATH_MAX);
		if (rs < 0)
			return -1;
		path[r] = 0;
		return symlink(path, dest);
	} else {
		return mknod(dest, st.st_mode, st.st_rdev);
	}
err:
	delete(dest, 1);
	return -1;
}

int move(char *src, char *dest) {
	if (rename(src, dest) &&
			(errno != EXDEV || copy(src, dest, 1) || delete(src, 1)))
		return -1;
	return 0;
}

struct file_list {
	struct file_list *next;
	int dir;
	char *filename;
	char *backup;
};

struct file_list *file_list_add(struct file_list *fl, int dir, char *tgtdir,
		char *filename) {
	struct file_list *new, *curr;
	for (curr = fl; curr; curr = curr->next) {
		switch (paths_equal(filename, curr->filename)) {
		case 1:
			return fl;
		case -1:
			target_emsg(tgtdir, filename, "out of memory");
			return NULL;
		}
	}
	new = malloc(sizeof(struct file_list));
	if (!new) {
		target_emsg(tgtdir, filename, "out of memory");
		return NULL;
	}
	new->filename = strdup(filename);
	if (!(new->filename)) {
		target_emsg(tgtdir, filename, "out of memory");
		free(new);
		return NULL;
	}
	new->backup = NULL;
	new->dir = dir;
	new->next = fl;
	return new;
}

char *backup_dir(char *tgtdir, char *name) {
	DIR *dir;
	struct dirent *de;
	char *tmp, *tmpdir, *src, *dest;
	size_t len, clen;
	len = strlen(name);
	tmp = emalloc(len + 13, name);
	if (!tmp)
		return NULL;
	sprintf(tmp, "%s-tspm.XXXXXX", name);
	tmpdir = mkdtemp(tmp);
	if (!tmpdir) {
		target_error(tgtdir, name);
		goto err;
	}
	dir = opendir(name);
	if (!dir) {
		target_error(tgtdir, name);
		goto err;
	}
	for (de = readdir(dir); de; de = readdir(dir)) {
		if (strcmp(de->d_name, ".") && strcmp(de->d_name, "..")) {
			clen = strlen(de->d_name);
			src = malloc(len + clen + 2);
			if (!src) {
				target_emsg(tgtdir, name, "out of memory");
				goto err2;
			}
			dest = malloc(len + clen + 14);
			if (!dest) {
				target_emsg(tgtdir, name, "out of memory");
				goto err3;
			}
			sprintf(src, "%s/%s", name, de->d_name);
			sprintf(dest, "%s/%s", tmpdir, de->d_name);
			if (move(src, dest)) {
				target_error(tgtdir, src);
				goto err4;
			}
		}
	}
	closedir(dir);
	return tmpdir;
err4:
	free(dest);
err3:
	free(src);
err2:
	closedir(dir);
err:
	free(tmp);
	return NULL;
}

char *backup_file(char *tgtdir, char *filename) {
	char *tmp;
	int fd;
	tmp = emalloc(strlen(filename) + 13, filename);
	if (!tmp)
		return NULL;
	sprintf(tmp, "%s-tspm.XXXXXX", filename);
	fd = mkstemp(tmp);
	if (fd < 0) {
		target_error(tgtdir, filename);
		free(tmp);
		return NULL;
	}
	close(fd);
	if (move(filename, tmp)) {
		target_error(tgtdir, filename);
		delete(tmp, 0);
		free(tmp);
		return NULL;
	}
	return tmp;
}

struct package_extractor {
	struct extractor e;
	char *tgtdir;
	struct file_list *oldfl;
	struct file_list *newfl;
	int force, overwrite;
};

int add_filename(struct package_extractor *pe, int dir, char *filename,
		char *backup) {
	struct file_list *fl;
	fl = file_list_add(pe->newfl, dir, pe->tgtdir, filename);
	if (!fl) {
		if (pe->overwrite) {
			if (backup) {
				if (move(backup, filename))
					target_error(pe->tgtdir, filename);
			} else {
				if (delete(filename, 1))
					target_error(pe->tgtdir, filename);
			}
		}
		return -1;
	}
	pe->newfl = fl;
	if (pe->overwrite) {
		for (fl = pe->oldfl; fl; fl = fl->next) {
			switch (paths_equal(fl->filename, filename)) {
			case 1:
				fl->backup = backup;
				break;
			case -1:
				target_emsg(pe->tgtdir, filename, "out of memory");
				return -1;
			}
		}
	}
	return 0;
}

int op_flags(struct package_extractor *pe, int track) {
	int flags;
	flags = EXTRACT_PERMISSION;
	if (!track || !pe->force || pe->overwrite)
		flags |= EXTRACT_KEEP;
	if (!pe->force && !pe->overwrite)
		flags |= EXTRACT_NOISYKEEP;
	return flags;
}

char *backup(struct package_extractor *pe, char *filename) {
	struct file_list *fl;
	struct stat st;
	for (fl = pe->oldfl; fl; fl = fl->next) {
		switch (paths_equal(fl->filename, filename)) {
		case 1:
			if (lstat(filename, &st))
				return NULL;
			if (S_ISDIR(st.st_mode))
				return backup_dir(pe->tgtdir, filename);
			else
				return backup_file(pe->tgtdir, filename);
		case -1:
			target_emsg(pe->tgtdir, filename, "out of memory");
			return NULL;
		}
	}
	return NULL;
}

char *backup_or_fail(struct package_extractor *pe, char *filename) {
	char *tmp;
	if (pe->overwrite) {
		tmp = backup(pe, filename);
		if (!tmp)
			return NULL;
		return tmp;
	} else {
		return NULL;
	}
}

int package_extractor_write(struct extractor *e, struct stream *in,
		char *filename, mode_t mode, time_t modtime, size_t size, int track,
		int reopen) {
	struct package_extractor *pe;
	char *tmp;
	int r;
	pe = (struct package_extractor *)e;
	tmp = NULL;
	r = ex_write(in, filename, mode, modtime, size, reopen,
		op_flags(pe, track), pe->tgtdir);
	if (r < 1)
		return r;
	if (r == 1) {
		if (track) {
			tmp = backup_or_fail(pe, filename);
			if (tmp || pe->force) {
				r = ex_write(in, filename, mode, modtime, size, reopen,
					EXTRACT_PERMISSION, pe->tgtdir);
			} else {
				target_emsg(pe->tgtdir, filename, "File already exists");
				x_stream_seek(in, size);
				return -1;
			}
		} else {
			x_stream_seek(in, size);
		}
	}
	if (r < 1)
		return r;
	if (track && add_filename(pe, 0, filename, tmp)) {
		if (r > 1 && !pe->overwrite)
			delete(filename, 0);
		return -1;
	}
	return 1;
}

int package_extractor_symlink(struct extractor *e, char *target,
		char *filename, int track) {
	struct package_extractor *pe;
	char *tmp;
	int r;
	pe = (struct package_extractor *)e;
	tmp = NULL;
	r = ex_symlink(target, filename, op_flags(pe, track), pe->tgtdir);
	if (r < 1)
		return r;
	if (r == 1) {
		if (track) {
			tmp = backup_or_fail(pe, filename);
			if (tmp || pe->force) {
				r = ex_symlink(target, filename, EXTRACT_PERMISSION,
					pe->tgtdir);
			} else {
				target_emsg(pe->tgtdir, filename, "File already exists");
				return -1;
			}
		}
	}
	if (r < 1)
		return r;
	if (track && add_filename(pe, 0, filename, tmp)) {
		if (r > 1 && !pe->overwrite)
			delete(filename, 0);
		return -1;
	}
	return 1;
}

int package_extractor_link(struct extractor *e, char *target, char *filename,
		int track) {
	struct package_extractor *pe;
	char *tmp;
	int r;
	pe = (struct package_extractor *)e;
	tmp = NULL;
	r = ex_link(target, filename, op_flags(pe, track), pe->tgtdir);
	if (r < 1)
		return r;
	if (r == 1) {
		if (track) {
			tmp = backup_or_fail(pe, filename);
			if (tmp || pe->force) {
				r = ex_link(target, filename, EXTRACT_PERMISSION, pe->tgtdir);
			} else {
				target_emsg(pe->tgtdir, filename, "File already exists");
				return -1;
			}
		}
	}
	if (r < 1)
		return r;
	/* one can not make a hard link to a directory */
	if (track && add_filename(pe, 0, filename, tmp)) {
		if (r > 1 && !pe->overwrite)
			delete(filename, 0);
		return -1;
	}
	return 1;
}

int package_extractor_mknod(struct extractor *e, char *filename, mode_t mode,
		dev_t dev, int track) {
	struct package_extractor *pe;
	char *tmp;
	int r;
	pe = (struct package_extractor *)e;
	tmp = NULL;
	r = ex_mknod(filename, mode, dev, op_flags(pe, track), pe->tgtdir);
	if (r < 1)
		return r;
	if (r == 1) {
		if (track) {
			tmp = backup_or_fail(pe, filename);
			if (tmp || pe->force) {
				r = ex_mknod(filename, mode, dev, EXTRACT_PERMISSION,
					pe->tgtdir);
			} else {
				target_emsg(pe->tgtdir, filename, "File already exists");
				return -1;
			}
		}
	}
	if (r < 1)
		return r;
	if (track && add_filename(pe, 0, filename, tmp)) {
		if (r > 1 && !pe->overwrite)
			delete(filename, 0);
		return -1;
	}
	return 1;
}

int package_extractor_mkdir(struct extractor *e, char *filename, mode_t mode,
		int track) {
	struct package_extractor *pe;
	int r, flags;
	pe = (struct package_extractor *)e;
	flags = EXTRACT_PERMISSION;
	if (!track || !pe->force)
		flags |= EXTRACT_KEEP;
	r = ex_mkdir(filename, mode, flags, pe->tgtdir);
	if (r < 1)
		return r;
	if (track && add_filename(pe, 1, filename, NULL)) {
		if (r > 1 && !pe->overwrite)
			delete(filename, 1);
		return -1;
	}
	return 1;
}

struct package_extractor *new_package_extractor(struct file_list *oldfl,
		char *tgtdir, int overwrite, int force) {
	struct package_extractor *pe;
	pe = malloc(sizeof(struct package_extractor));
	if (!pe) {
		fprintf(stderr, "%s: out of memory\n");
		return NULL;
	}
	pe->e.write = package_extractor_write;
	pe->e.symlink = package_extractor_symlink;
	pe->e.link = package_extractor_link;
	pe->e.mknod = package_extractor_mknod;
	pe->e.mkdir = package_extractor_mkdir;
	pe->tgtdir = tgtdir;
	pe->oldfl = oldfl;
	pe->newfl = NULL;
	pe->overwrite = overwrite;
	pe->force = force;
	return pe;
}

void rollback(struct package_extractor *pe) {
	struct file_list *fl;
	for (fl = pe->newfl; fl; fl = fl->next)
		if (strcmp(fl->filename, "."))
			delete(fl->filename, 0);
}

int rollback_for_overwrite(FILE *f, struct package_extractor *pe) {
	struct file_list *fl, *ofl;
	for (fl = pe->newfl; fl; fl = fl->next) {
		if (strcmp(fl->filename, ".")) {
			for (ofl = pe->oldfl; ofl; ofl = ofl->next) {
				if (paths_equal(fl->filename, ofl->filename) == 1) {
					if (ofl->backup && move(ofl->backup, fl->filename))
						target_error(pe->tgtdir, fl->filename);
					else
						goto rbkdone;
				}
			}
			if (strcmp(fl->filename, "."))
				delete(fl->filename, 0);
rbkdone:
			continue;
		}
	}
	for (fl = pe->oldfl; fl; fl = fl->next) {
		if (fl->dir) {
			if (write_kvp(f, "Directory", fl->filename))
				return -1;
		} else {
			if (write_kvp(f, "File", fl->filename))
				return -1;
		}
	}
	return 0;
}

int cleanup_overwrite(struct package_extractor *pe) {
	struct file_list *fl;
	char *name;
	int result;
	result = 0;
	for (fl = pe->oldfl; fl; fl = fl->next) {
		if (fl->backup)
			name = fl->backup;
		else
			name = fl->filename;
		if (strcmp(name, ".") && delete(name, 0) && errno != ENOTEMPTY &&
				errno != EBUSY) {
			target_error(pe->tgtdir, name);
			result = -1;
		}
	}
	return result;
}

void read_package_header(struct stream *in, char **name, char **vers,
		struct dep **deps) {
	struct dep *new, *cd;
	char buf[512];
	char namebuf[101];
	char verbuf[25];
	char *depdata;
	size_t size, tarsize, i;
	if (x_stream_read(in, buf, 512) < 512 || buf[156] != '#') {
		fprintf(stderr, "%s: %s: not a TSPM package\n", progname,
			in->filename);
		exit(1);
	}
	if (!checksum_is_valid(buf, in->filename))
		exit(1);
	size = strtol(buf + 124, NULL, 8);
	tarsize = ((size + 511) & ~511);
	depdata = xmalloc(tarsize, in->filename);
	cd = NULL;
	if (x_stream_read(in, depdata, tarsize) < tarsize)
		exit(1);
	*deps = NULL;
	for (i = 0; i < size;) {
		new = xmalloc(sizeof(struct dep), in->filename);
		new->name = strdup(depdata + i);
		if (!new->name)
			goto oom;
		if (strchr(new->name, 10)) {
			fprintf(stderr,
				"%s: invalid TSP file: dependency name contains newline\n",
				progname);
			exit(1);
		}
		i += strlen(new->name) + 1;
		new->min_version = strdup(depdata + i);
		if (!new->min_version)
			goto oom;
		if (strchr(new->min_version, 10)) {
			fprintf(stderr,
				"%s: invalid TSP file: dependency version contains newline\n",
				progname);
			exit(1);
		}
		i += strlen(new->min_version) + 1;
		new->max_version = strdup(depdata + i);
		if (!new->max_version)
			goto oom;
		if (strchr(new->max_version, 10)) {
			fprintf(stderr,
				"%s: invalid TSP file: dependency version contains newline\n",
				progname);
			exit(1);
		}
		i += strlen(new->max_version) + 1;
		new->next = NULL;
		if (!cd)
			*deps = new;
		else
			cd->next = new;
		cd = new;
	}
	memcpy(namebuf, buf, 100);
	namebuf[100] = 0;
	*name = strdup(namebuf);
	if (!*name)
		goto oom;
	if (strchr(*name, 10)) {
		fprintf(stderr,
			"%s: invalid TSP file: package name contains newline\n", progname);
		exit(1);
	}
	memcpy(verbuf, buf + 100, 24);
	verbuf[24] = 0;
	*vers = strdup(verbuf);
	if (!*vers)
		goto oom;
	if (strchr(*vers, 10)) {
		fprintf(stderr,
			"%s: invalid TSP file: package name contains newline\n", progname);
		exit(1);
	}
	return;
oom:
	fprintf(stderr, "%s: %s: out of memory\n", progname, in->filename);
	exit(1);
}

void overwrite_dependency_check(char *name, char *vers, char *depname,
		char *target) {
	char *datafile, *key, *value;
	FILE *f;
	int found;
	found = 0;
	if (strcmp(name, depname)) {
		datafile = malloc(10 + strlen(depname));
		if (!datafile) {
			fprintf(stderr, "%s: out of memory\n", progname);
			exit(1);
		}
		sprintf(datafile, "etc/tspm/%s", depname);
		f = fopen(datafile, "r");
		if (!f)
			target_die(target, datafile);
		while (read_kvp(f, datafile, &key, &value)) {
			if (!strcmp(key, "Dependency")) {
				if (strcmp(value, name)) {
					if (found)
						goto cleanup;
				} else {
					found = 1;
				}
			} else if (!strcmp(key, "Minimum") && found) {
				if (strverscmp(vers, value) < 0) {
					fprintf(stderr,
						"%s: package %s depends on at least version %s of "
						"package %s (%s would be installed)\n", progname,
						depname, value, name, vers);
					exit(1);
				}
			} else if (!strcmp(key, "Maximum") && found) {
				if (strverscmp(vers, value) > 0) {
					fprintf(stderr,
						"%s: package %s depends on at most version %s of "
						"package %s (%s would be installed)\n", progname,
						depname, value, name, vers);
					exit(1);
				}
			}
		}
cleanup:
		fclose(f);
		free(datafile);
	}
}

void dependency_check(struct dep *dep, char *target) {
	char *datafile, *key, *value;
	FILE *f;
	datafile = malloc(10 + strlen(dep->name));
	if (!datafile) {
		fprintf(stderr, "%s: out of memory\n", progname);
		exit(1);
	}
	sprintf(datafile, "etc/tspm/%s", dep->name);
	f = fopen(datafile, "r");
	if (!f) {
		if (errno == ENOENT) {
			fprintf(stderr, "%s: dependency %s not installed\n", progname,
				dep->name);
			exit(1);
		} else {
			target_die(target, datafile);
		}
	}
	while (read_kvp(f, datafile, &key, &value)) {
		if (!strcmp(key, "Version")) {
			if (dep->min_version && *(dep->min_version) &&
					strverscmp(value, dep->min_version) < 0) {
				fprintf(stderr,
					"%s: dependency %s must be at least version "
					"%s (%s installed)\n", progname, dep->name,
					dep->min_version, value);
				exit(1);
			}
			if (dep->max_version && *(dep->max_version) &&
					strverscmp(value, dep->max_version) > 0) {
				fprintf(stderr,
					"%s: dependency %s must be at most version "
					"%s (%s installed)\n", progname, dep->name,
					dep->max_version, value);
				exit(1);
			}
			fclose(f);
			return;
		}
		free(key);
		free(value);
	}
	fclose(f);
	if ((dep->max_version && *(dep->max_version)) ||
			(dep->min_version && *(dep->min_version))) {
		fprintf(stderr,
			"%s: internal record store corrupted: can not determine version "
			"of dependency %s\n", progname, dep->name);
		exit(1);
	} else {
		fprintf(stderr,
			"%s: warning: internal record store corrupted: can not determine "
			"version of dependency %s\n", progname, dep->name);
	}
}

int main(int argc, char **argv) {
	struct dirent *de;
	struct package_extractor *pe;
	struct extractor *e;
	struct stream *file;
	struct dep *deps, *cd;
	struct file_list *fl, *oldfl;
	DIR *d;
	FILE *f;
	char *name, *vers, *datafile, *filename, *key, *value, *target;
	int r, fd, i, j, overwrite, iinc, force;
	progname = argv[0];
	filename = NULL;
	target = "/";
	overwrite = 0;
	force = 0;
	for (i = 1; i < argc; i += iinc + 1) {
		iinc = 0;
		if (argv[i][0] == '-' && argv[i][1]) {
			for (j = 1; argv[i][j]; j++) {
				switch (argv[i][j]) {
				case 'f':
					force = 1;
					break;
				case 'o':
					overwrite = 1;
					break;
				case 't':
					if (i + iinc + 1 >= argc || argv[i+iinc+1][0] == '-') {
						fprintf(stderr, "%s: -t requires argument\n",
							progname);
						exit(1);
					}
					iinc++;
					target = argv[i + iinc];
					break;
				case 'V':
					version("inspak");
					return 0;
				case '?':
					version("inspak");
					fprintf(stderr,
						"\n%s [-of] [-t TARGET] FILENAME\n"
						"\tInstall a package from the package file named\n"
						"\tFILENAME. If -o is specified, replace any package\n"
						"\talready installed with the same name. If -f is\n"
						"\tspecified, overwrite any files already present\n"
						"\twith those from the package. If -t is specified,\n"
						"\ttreat TARGET as the root directory.\n"
						"%s -?\n"
						"\tShow this usage message.\n"
						"%s -V\n"
						"\tDisplay the version of this command.\n",
						progname, progname, progname);
					return 0;
				default:
					fprintf(stderr, "%s: invalid -%c\n", progname, argv[i][j]);
					exit(1);
				}
			}
		} else {
			if (filename) {
				fprintf(stderr, "%s: invalid %s: too many filenames\n",
					progname, argv[i]);
				exit(1);
			}
			filename = argv[i];
		}
	}
	if (!filename) {
		fprintf(stderr, "%s: filename required\n", progname);
		exit(1);
	}
	file = xz_stream_open(filename, "r");
	if (!file)
		exit(1);
	read_package_header(file, &name, &vers, &deps);
	if (chdir(target))
		die(target);
	if (overwrite) {
		d = opendir("etc/tspm");
		if (!d) {
			if (errno != ENOENT)
				target_die(target, "etc/tspm");
		} else {
			errno = 0;
			for (de = readdir(d); de; de = readdir(d)) {
				overwrite_dependency_check(name, vers, de->d_name, target);
				errno = 0;
			}
			if (errno)
				target_die(target, "etc/tspm");
			closedir(d);
		}
	}
	for (cd = deps; cd; cd = cd->next)
		dependency_check(cd, target);
	datafile = malloc(10 + strlen(name));
	if (!datafile) {
		fprintf(stderr, "%s: out of memory\n", progname);
		return 1;
	}
	sprintf(datafile, "etc/tspm/%s", name);
	errno = 0;
	fd = open(datafile, O_WRONLY|O_CREAT|O_NOFOLLOW|O_EXCL|O_TRUNC, 0644);
	if (fd < 0 && errno == ENOENT) {
		if (overwrite) {
			fprintf(stderr, "%s: package %s not installed\n", progname, name);
			exit(1);
		} else {
			errno = 0;
			r = mkdir("etc/tspm", 0755);
			if (r && errno == ENOENT) {
				if (mkdir("etc", 0755))
					target_die(target, "etc");
				r = mkdir("etc/tspm", 0755);
			}
			if (r)
				target_die(target, "etc/tspm");
			fd = open(datafile, O_WRONLY|O_CREAT|O_NOFOLLOW|O_EXCL|O_TRUNC,
				0644);
		}
	}
	oldfl = NULL;
	if (fd < 0) {
		if (errno == EEXIST) {
			if (overwrite) {
				f = fopen(datafile, "r");
				if (!f)
					target_die(target, datafile);
				while (read_kvp(f, datafile, &key, &value)) {
					if (!strcmp(key, "Directory")) {
						fl = file_list_add(oldfl, 1, target, value);
						if (!fl)
							exit(1);
						oldfl = fl;
					} else if (!strcmp(key, "File"))  {
						fl = file_list_add(oldfl, 0, target, value);
						if (!fl)
							exit(1);
						oldfl = fl;
					}
					free(key);
					free(value);
				}
				fclose(f);
				fd = open(datafile, O_WRONLY|O_NOFOLLOW|O_TRUNC, 0644);
				if (fd < 0)
					target_die(target, datafile);
			} else {
				fprintf(stderr, "%s: package %s already installed\n", progname,
					name);
				exit(1);
			}
		} else {
			target_die(target, datafile);
		}
	}
	f = fdopen(fd, "w");
	if (!f)
		target_die(target, datafile);
	pe = new_package_extractor(oldfl, target, overwrite, force);
	e = (struct extractor *)pe;
	for (r = extract_file(e, target, file, 1); r;
			r = extract_file(e, target, file, 1))
		if (r < 0)
			goto rbk;
	file->close(file);
	if (write_kvp(f, "Version", vers))
		goto out_err;
	for (cd = deps; cd; cd = cd->next) {
		if (cd->name && *(cd->name)) {
			if (write_kvp(f, "Dependency", cd->name))
				goto out_err;
			if (cd->min_version && *(cd->min_version) &&
					write_kvp(f, "Minimum", cd->min_version))
				goto out_err;
			if (cd->max_version && *(cd->max_version) &&
					write_kvp(f, "Maximum", cd->max_version))
				goto out_err;
		}
	}
	fl = pe->newfl;
	for (; fl; fl = fl->next) {
		if (fl->dir) {
			if (write_kvp(f, "Directory", fl->filename))
				goto out_err;
		} else {
			if (write_kvp(f, "File", fl->filename))
				goto out_err;
		}
	}
	fclose(f);
	if (overwrite && cleanup_overwrite(pe))
		return 1;
	return 0;
out_err:
	target_error(target, datafile);
rbk:
	if (overwrite) {
		if (rollback_for_overwrite(f, pe))
			target_error(target, datafile);
	} else {
		rollback(pe);
	}
	fclose(f);
	if (!overwrite && delete(datafile, 0))
		target_error(target, datafile);
	return 1;
}
