#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>

#include <zlib.h>
#include <bzlib.h>
#include <lzma.h>

#include "lib.h"

char *progname;

void *emalloc(size_t size, char *filename) {
	void *result;
	result = malloc(size);
	if (!result)
		fprintf(stderr, "%s: %s: out of memory\n", progname, filename);
	return result;
}

void *xmalloc(size_t size, char *filename) {
	void *result;
	result = emalloc(size, filename);
	if (!result)
		exit(1);
	return result;
}

void *erealloc(void *buf, size_t size, char *filename) {
	void *result;
	result = realloc(buf, size * 2);
	if (!result)
		fprintf(stderr, "%s: %s: out of memory\n", progname, filename);
	return result;
}

void *xrealloc(void *buf, size_t size, char *filename) {
	void *result;
	result = erealloc(buf, size, filename);
	if (!result)
		exit(1);
	return result;
}

struct file_stream {
	struct stream s;
	FILE *f;
};

struct gz_stream {
	struct stream s;
	gzFile f;
	int err;
};

struct bz2_stream {
	struct stream s;
	BZFILE *f;
	int for_write;
	int end;
	int err;
};

struct xz_stream {
	struct stream s;
	lzma_stream ls;
	struct stream *fs;
	int for_write;
	int err;
	char *in;
	size_t in_pos;
	size_t in_fill;
	size_t in_size;
	char *out;
	size_t out_pos;
	size_t out_fill;
	size_t out_size;
};

void error(char *filename) {
	fprintf(stderr, "%s: %s: %s\n", progname, filename, strerror(errno));
	errno = 0;
}

void die(char *filename) {
	error(filename);
	exit(1);
}

void gz_stream_emsg(struct gz_stream *stream) {
	fprintf(stderr, "%s: %s: %s\n", progname, stream->s.filename,
		gzerror(stream->f, NULL));
}

void bz2_stream_emsg(char *filename, int err) {
	switch (err) {
	case BZ_MEM_ERROR:
		fprintf(stderr, "%s: %s: out of memory\n", progname, filename);
		break;
	case BZ_DATA_ERROR:
		fprintf(stderr, "%s: %s: invalid BZ2 checksum\n", progname, filename);
		break;
	case BZ_DATA_ERROR_MAGIC:
		fprintf(stderr, "%s: %s: invalid BZ2 signature\n", progname, filename);
		break;
	case BZ_IO_ERROR:
		error(filename);
		break;
	case BZ_UNEXPECTED_EOF:
		fprintf(stderr, "%s: %s: short read\n", progname, filename);
		break;
	default:
		fprintf(stderr, "%s: %s: unexpected error %d\n", progname, filename,
			err);
	}
}

void xz_stream_emsg(char *filename, int result) {
	switch (result) {
	case LZMA_MEM_ERROR:
		fprintf(stderr, "%s: %s: out of memory\n", progname, filename);
		break;
	case LZMA_DATA_ERROR:
	case LZMA_FORMAT_ERROR:
		fprintf(stderr, "%s: %s: invalid XZ checksum\n", progname, filename);
		break;
	default:
		fprintf(stderr, "%s: %s: unexpected error %d\n", progname, filename,
			result);
	}
}

size_t file_stream_read(struct stream *stream, char *buf, size_t size) {
	struct file_stream *fs;
	size_t result;	
	fs = (struct file_stream *)stream;
	result = fread(buf, 1, size, fs->f);
	if (ferror(fs->f))
		error(fs->s.filename);
	return result;
}

size_t gz_stream_read(struct stream *stream, char *buf, size_t size) {
	struct gz_stream *gs;
	size_t result;
	int errnum;
	gs = (struct gz_stream *)stream;
	result = gzread(gs->f, buf, size);
	gzerror(gs->f, &errnum);
	if (errnum < 0) {
		gz_stream_emsg(gs);
		gs->err = 1;
	}
	return result;
}

size_t bz2_stream_read(struct stream *stream, char *buf, size_t size) {
	struct bz2_stream *bs;
	size_t result;
	int err;
	bs = (struct bz2_stream *)stream;
	if (bs->end)
		return 0;
	result = BZ2_bzRead(&err, bs->f, buf, size);
	if (err == BZ_STREAM_END) {
		bs->end = 1;
		return result;
	} else if (err != BZ_OK) {
		bz2_stream_emsg(bs->s.filename, err);
		bs->err = 1;
		return 0;
	} else {
		return result;
	}
}

int xz_stream_fill_buffers(struct xz_stream *xs) {
	lzma_ret result;
	size_t i;
	if (xs->in_pos == xs->in_fill) {
		xs->in_pos = 0;
		xs->in_fill = xs->fs->read(xs->fs, xs->in, xs->in_size);
		if (!xs->in_fill)
			return 0;
	}
	xs->out_pos = 0;
	xs->out_fill = 0;
	xs->ls.next_in = xs->in + xs->in_pos;
	xs->ls.avail_in = xs->in_fill - xs->in_pos;
	xs->ls.total_in = 0;
	xs->ls.next_out = xs->out;
	xs->ls.avail_out = xs->out_size;
	xs->ls.total_out = 0;
	result = lzma_code(&(xs->ls), LZMA_RUN);
	if (result != LZMA_OK) {
		xz_stream_emsg(xs->s.filename, result);
		xs->err = 1;
		return -1;
	}
	xs->out_fill = xs->ls.total_out;
	xs->in_pos += xs->ls.total_in;
	return 1;
}

size_t xz_stream_read(struct stream *stream, char *buf, size_t size) {
	struct xz_stream *xs;
	size_t i;
	xs = (struct xz_stream *)stream;
	for (i = 0; i < size;) {
		if (xs->out_fill - xs->out_pos < size - i) {
			memcpy(buf + i, xs->out + xs->out_pos, xs->out_fill - xs->out_pos);
			i += xs->out_fill - xs->out_pos;
			xs->out_pos = xs->out_fill;
			if (xz_stream_fill_buffers(xs) <= 0)
				return i;
		} else {
			memcpy(buf + i, xs->out + xs->out_pos, size - i);
			xs->out_pos += size - i;
			i = size;
		}
	}
	return i;
}

int file_stream_write(struct stream *stream, char *buf, size_t size) {
	struct file_stream *fs;
	size_t result;
	if (size) {
		fs = (struct file_stream *)stream;
		result = fwrite(buf, 1, size, fs->f);
		if (result < size) {
			error(fs->s.filename);
			return -1;
		}
	}
	return 0;
}

int gz_stream_write(struct stream *stream, char *buf, size_t size) {
	struct gz_stream *gs;
	gs = (struct gz_stream *)stream;
	if (gzwrite(gs->f, buf, size) < size) {
		gz_stream_emsg(gs);
		gs->err = 1;
		return -1;
	} else {
		return 0;
	}
}

int bz2_stream_write(struct stream *stream, char *buf, size_t size) {
	struct bz2_stream *bs;
	int err;
	bs = (struct bz2_stream *)stream;
	BZ2_bzWrite(&err, bs->f, buf, size);
	if (err != BZ_OK) {
		bz2_stream_emsg(bs->s.filename, err);
		bs->err = 1;
		return -1;
	} else {
		return 0;
	}
}

lzma_ret xz_stream_compress_and_write(struct xz_stream *xs, int *write_result,
		lzma_action action) {
	lzma_ret result;
	xs->ls.next_out = xs->out;
	xs->ls.avail_out = xs->out_size;
	xs->ls.total_out = 0;
	result = lzma_code(&(xs->ls), action);
	*write_result = xs->fs->write(xs->fs, xs->out, xs->ls.total_out);
	return result;
}

int xz_stream_write(struct stream *stream, char *buf, size_t size) {
	struct xz_stream *xs;
	lzma_ret result;
	size_t i, j, esize, osize;
	int write_result;
	xs = (struct xz_stream *)stream;
	for (i = 0; i < size; i += osize) {
		if (i + xs->in_size < size + xs->in_pos)
			esize = xs->in_size;
		else
			esize = size + xs->in_pos - i;
		osize = esize - xs->in_pos;
		memcpy(xs->in + xs->in_pos, buf + i, osize);
		xs->ls.next_in = xs->in;
		xs->ls.avail_in = esize;
		xs->ls.total_in = 0;
		result = xz_stream_compress_and_write(xs, &write_result, LZMA_RUN);
		if (result != LZMA_OK) {
			return -1;
		} else if (write_result) {
			xz_stream_emsg(xs->s.filename, result);
			xs->err = 1;
			return -1;
		}
		xs->in_pos = xs->ls.avail_in;
		if (xs->in_pos && xs->ls.total_in)
			memmove(xs->in, xs->in + xs->ls.total_in, xs->in_pos);
	}
	return 0;
}

void file_stream_close(struct stream *stream) {
	struct file_stream *fs;
	fs = (struct file_stream *)stream;
	fclose(fs->f);
	free(fs);
}

void gz_stream_close(struct stream *stream) {
	struct gz_stream *gs;
	gs = (struct gz_stream *)stream;
	gzclose(gs->f);
	free(gs);
}

void bz2_stream_close(struct stream *stream) {
	struct bz2_stream *bs;
	int err;
	bs = (struct bz2_stream *)stream;
	if (bs->for_write)
		BZ2_bzWriteClose(&err, bs->f, 0, NULL, NULL);
	else
		BZ2_bzReadClose(&err, bs->f);
	free(bs);
}

void xz_stream_close(struct stream *stream) {
	struct xz_stream *xs;
	lzma_ret result;
	int write_result;
	xs = (struct xz_stream *)stream;
	if (xs->for_write) {
		do {
			result = xz_stream_compress_and_write(xs, &write_result,
				LZMA_FINISH);
			if (write_result) {
				error(xs->s.filename);
				goto cleanup;
			} else if (result != LZMA_OK && result != LZMA_STREAM_END) {
				xz_stream_emsg(xs->s.filename, result);
				goto cleanup;
			}
		} while (result != LZMA_STREAM_END);
	}
	lzma_end(&(xs->ls));
cleanup:
	xs->fs->close(xs->fs);
	if (xs->out)
		free(xs->out);
	if (xs->in)
		free(xs->in);
	free(xs);
}

int file_stream_error(struct stream *stream) {
	struct file_stream *fs;
	fs = (struct file_stream *)stream;
	return !!ferror(fs->f);
}

int gz_stream_error(struct stream *stream) {
	struct gz_stream *gs;
	gs = (struct gz_stream *)stream;
	return gs->err;
}

int bz2_stream_error(struct stream *stream) {
	struct bz2_stream *bs;
	bs = (struct bz2_stream *)stream;
	return bs->err;
}

int xz_stream_error(struct stream *stream) {
	struct xz_stream *xs;
	xs = (struct xz_stream *)stream;
	return xs->err;
}

struct stream *file_stream_new(FILE *f, char *filename) {
	struct file_stream *result;
	result = emalloc(sizeof(struct file_stream), filename);
	if (!result)
		return NULL;
	result->s.read = file_stream_read;
	result->s.write = file_stream_write;
	result->s.close = file_stream_close;
	result->s.error = file_stream_error;
	result->s.filename = filename;
	result->f = f;
	return (struct stream *)result;
}

struct stream *gz_stream_new(gzFile f, char *filename) {
	struct gz_stream *result;
	result = emalloc(sizeof(struct gz_stream), filename);
	if (!result)
		return NULL;
	result->s.read = gz_stream_read;
	result->s.write = gz_stream_write;
	result->s.close = gz_stream_close;
	result->s.error = gz_stream_error;
	result->s.filename = filename;
	result->f = f;
	result->err = 0;
	return (struct stream *)result;
}

struct stream *bz2_stream_new(BZFILE *f, char *filename, int for_write) {
	struct bz2_stream *result;
	result = emalloc(sizeof(struct bz2_stream), filename);
	if (!result)
		return NULL;
	result->s.read = bz2_stream_read;
	result->s.write = bz2_stream_write;
	result->s.close = bz2_stream_close;
	result->s.error = bz2_stream_error;
	result->s.filename = filename;
	result->for_write = for_write;
	result->end = 0;
	result->err = 0;
	result->f = f;
	return (struct stream *)result;
}

struct stream *xz_stream_new(struct stream *fs, char *filename,
		int for_write, int level) {
	struct xz_stream *result;
	lzma_ret r;
	lzma_stream tmp = LZMA_STREAM_INIT;
	result = emalloc(sizeof(struct xz_stream), filename);
	if (!result)
		return NULL;
	result->s.read = xz_stream_read;
	result->s.write = xz_stream_write;
	result->s.close = xz_stream_close;
	result->s.error = xz_stream_error;
	result->s.filename = filename;
	result->fs = fs;
	memcpy(&(result->ls), &tmp, sizeof(lzma_stream));
	result->for_write = for_write;
	if (for_write) {
		result->out_size = lzma_stream_buffer_bound(1024);
		result->in_size = 1024;
		r = lzma_easy_encoder(&(result->ls), level, LZMA_CHECK_CRC64);
	} else {
		result->out_size = 1024;
		result->in_size = 1024;
		r = lzma_stream_decoder(&(result->ls), UINT64_MAX, LZMA_CONCATENATED);
	}
	if (r != LZMA_OK) {
		xz_stream_emsg(filename, r);
		free(result);
		return NULL;
	}
	result->out = xmalloc(result->out_size, filename);
	result->in = xmalloc(result->in_size, filename);
	result->out_fill = 0;
	result->out_pos = 0;
	result->in_fill = 0;
	result->in_pos = 0;
	result->err = 0;
	return (struct stream *)result;
}

FILE *filename_adjust_open(char **filename, char *mode) {
	FILE *result;
	result = NULL;
	if (!strcmp(*filename, "-")) {
		if (!strcmp(mode, "w")) {
			result = stdout;
			*filename = "(standard output)";
		} else if (!strcmp(mode, "r")) {
			result = stdin;
			*filename = "(standard input)";
		}
	} else {
		result = fopen(*filename, mode);
		if (!result) {
			error(*filename);
			return NULL;
		}
	}
	return result;
}

struct stream *file_stream_open(char *filename, char *mode) {
	FILE *f;
	char *rename;
	struct stream *result;
	rename = filename;
	f = filename_adjust_open(&rename, mode);
	if (!f)
		return NULL;
	result = file_stream_new(f, rename);
	if (!result)
		fclose(f);
	return result;
}

gzFile gz_dup_dopen(int fd, char *filename, char *mode) {
	gzFile result;
	int newfd;
	newfd = dup(fd);
	if (newfd < 0) {
		error(filename);
		return NULL;
	}
	result = gzdopen(newfd, mode);
	if (!result) {
		error(filename);
		close(newfd);
		return NULL;
	}
	return result;
}

struct stream *gz_stream_open(char *filename, char *mode) {
	gzFile f;
	struct stream *result;
	if (!strcmp(filename, "-")) {
		if (strchr(mode, 'w')) {
			f = gz_dup_dopen(1, filename, mode);
			if (!f)
				return NULL;
		} else if (strchr(mode, 'r')) {
			f = gz_dup_dopen(0, filename, mode);
			if (!f)
				return NULL;
		}
	} else {
		f = gzopen(filename, mode);
		if (!f) {
			error(filename);
			return NULL;
		}
	}
	result = gz_stream_new(f, filename);
	if (!result)
		gzclose(f);
	return result;
}

struct stream *gz_compress_dopen(int fd, char *filename, int level) {
	gzFile f;
	struct stream *result;
	char mode[3];
	mode[0] = 'w';
	mode[1] = '0' + level;
	mode[2] = 0;
	f = gz_dup_dopen(fd, filename, mode);
	if (!f)
		return NULL;
	result = gz_stream_new(f, filename);
	if (!result)
		gzclose(f);
	return result;
}

struct stream *bz2_stream_from_file_ex(FILE *f, char *filename, int for_write,
		int level) {
	BZFILE *b;
	struct stream *result;
	int err;
	if (for_write)
		b = BZ2_bzWriteOpen(&err, f, level, 0, 30);
	else
		b = BZ2_bzReadOpen(&err, f, 0, 0, NULL, 0);
	if (err != BZ_OK) {
		bz2_stream_emsg(filename, err);
		return NULL;
	}
	result = bz2_stream_new(b, filename, for_write);
	if (!result) {
		if (for_write)
			BZ2_bzWriteClose(&err, b, 0, NULL, NULL);
		else
			BZ2_bzReadClose(&err, b);
	}
	return result;
}

struct stream *bz2_stream_from_file(FILE *f, char *filename, int for_write) {
	return bz2_stream_from_file_ex(f, filename, for_write, 9);
}

struct stream *bz2_stream_open(char *filename, char *mode) {
	FILE *f;
	BZFILE *b;
	struct stream *result;
	char *rename;
	int err;
	rename = filename;
	f = filename_adjust_open(&rename, mode);
	if (!f)
		return NULL;
	result = bz2_stream_from_file(f, rename, !!(long)strchr(mode, 'w'));
	if (!result)
		fclose(f);
	return result;
}

struct stream *bz2_compress_dopen(int fd, char *filename, int level) {
	FILE *f;
	struct stream *result;
	f = fdopen(fd, "w");
	if (!f) {
		error(filename);
		return NULL;
	}
	result = bz2_stream_from_file_ex(f, filename, 1, level);
	if (!result)
		fclose(f);
	return result;
}

struct stream *xz_stream_open(char *filename, char *mode) {
	struct stream *s, *result;
	s = file_stream_open(filename, mode);
	if (!s)
		return NULL;
	result = xz_stream_new(s, filename, *mode == 'w', 6);
	if (!result)
		s->close(s);
	return result;
}

struct stream *xz_compress_dopen(int fd, char *filename, int level) {
	FILE *f;
	struct stream *s, *result;
	f = fdopen(fd, "w");
	if (!f) {
		error(filename);
		return NULL;
	}
	s = file_stream_new(f, filename);
	if (!s)
		goto err;
	result = xz_stream_new(s, filename, 1, level);
	if (!result)
		goto err2;
	return result;
err2:
	s->close(s);
err:
	fclose(f);
	return NULL;
}

size_t x_stream_read(struct stream *s, char *buf, size_t size) {
	size_t result;
	result = s->read(s, buf, size);
	if (s->error(s))
		exit(1);
	return result;
}

void x_stream_write(struct stream *s, char *buf, size_t size) {
	if (s->write(s, buf, size))
		exit(1);
}

int stream_seek(struct stream *stream, long offset) {
	char buf[1024];
	long i;
	for (i = 0; i < offset; i += 1024) {
		if (i + 1024 < offset) {
			if (stream->read(stream, buf, 1024) < 1024) {
				if (!stream->error(stream))
					fprintf(stderr, "%s: %s: short read\n", progname,
						stream->filename);
				return -1;
			}
		} else {
			if (stream->read(stream, buf, offset - i) + i < offset) {
				if (!stream->error(stream))
					fprintf(stderr, "%s: %s: short read\n", progname,
						stream->filename);
				return -1;
			}
		}
	}
	return 0;
}

void x_stream_seek(struct stream *s, long offset) {
	if (stream_seek(s, offset))
		exit(1);
}

size_t copy_all(struct stream *in, struct stream *out) {
	char buf[1024];
	size_t size, total;
	total = 0;
	for (size = in->read(in, buf, 1024); size;
			size = in->read(in, buf, 1024)) {
		if (out->write(out, buf, size))
			return total;
		total += size;
	}
	return total;
}

size_t x_copy_all(struct stream *in, struct stream *out) {
	size_t result;
	result = copy_all(in, out);
	if (in->error(in) || out->error(out))
		exit(1);
	return result;
}

size_t copy_bytes(struct stream *in, struct stream *out, size_t size) {
	char buf[1024];
	size_t i, r, cs;
	for (i = 0; i < size; i += 1024) {
		if (i + 1024 < size)
			cs = 1024;
		else
			cs =  size - i;
		r = x_stream_read(in, buf, cs);
		x_stream_write(out, buf, r);
		if (r < cs)
			return i + r;
	}
	return size;
}

void fill_bytes(struct stream *out, size_t size) {
	char buf[1024];
	size_t i;
	for (i = 0; i < 1024; i++)
		buf[i] = 0;
	for (i = 0; i < size; i += 1024) {
		if (i + 1024 < size)
			x_stream_write(out, buf, 1024);
		else
			x_stream_write(out, buf, size - i);
	}
}

char *construct_child_path(char *parent, char *child) {
	char *result;
	result = malloc(strlen(parent) + strlen(child) + 2);
	if (!result)
		return NULL;
	sprintf(result, "%s/%s", parent, child);
	return result;
}

int path_is_safe(char *path) {
	size_t len;
	len = strlen(path);
	return *path != '/' && !strstr(path, "/../") &&
		(len < 3 || strcmp("/..", path + len - 3)) && strcmp("..", path) &&
		strncmp("../", path, 3);
}
