#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <pwd.h>
#include <grp.h>
#include <libgen.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/sysmacros.h>

#include "lib.h"
#include "common.h"
#include "libtar.h"

void output_file(struct stream *in, struct stream *out) {
	fill_bytes(out, (512 - x_copy_all(in, out) % 512) % 512);
}

int checksum(char *block, int sgn) {
	size_t i;
	int result;
	result = 0;
	for (i = 0; i < 512; i++) {
		if (i >= 148 && i < 156)
			result += 32;
		else if (sgn)
			result = (result + (char)(block[i])) & 0x3ffff;
		else
			result = (result + (unsigned char)(block[i])) & 0x3ffff;
	}
	return result;
}

void output_infoblock(struct stream *out, char *path, mode_t mode, uid_t uid,
		gid_t gid, off_t size, time_t time, char block_type, char *target,
		char *username, char *groupname, dev_t dev) {
	unsigned char block[512];
	size_t i;
	for (i = 0; i < 512; i++)
		block[i] = 0;
	strncpy(block, path, 100);
	snprintf(block + 100, 8, "%07o", mode);
	snprintf(block + 108, 8, "%07o", uid);
	snprintf(block + 116, 8, "%07o", gid);
	snprintf(block + 124, 12, "%011o", size);
	snprintf(block + 136, 12, "%011o", time);
	block[156] = block_type;
	if (target)
		strncpy(block + 157, target, 100);
	sprintf(block + 257, "ustar");
	sprintf(block + 263, "00");
	if (username)
		strncpy(block + 265, username, 32);
	if (groupname)
		strncpy(block + 297, groupname, 32);
	snprintf(block + 329, 8, "%06o ", major(dev));
	snprintf(block + 337, 8, "%06o ", minor(dev));
	snprintf(block + 148, 8, "%06o ", checksum(block, 0));
	x_stream_write(out, block, 512);
}

void output_long_link(struct stream *out, char *path) {
	size_t i, len;
	len = strlen(path);
	output_infoblock(out, "././@LongLink", 0, 0, 0, strlen(path), 0, 'L', NULL,
		"root", "root", 0);
	x_stream_write(out, path, len);
	fill_bytes(out, (512 - len % 512) % 512);
}

int prepare_directory(struct path_info **pi, char *name, int linkfirst) {
	DIR *dir;
	struct dirent *de;
	int result;
	result = 0;
	dir = opendir(name);
	if (!dir) {
		error(name);
		return -1;
	}
	for (de = readdir(dir); de; de = readdir(dir))
		if (strcmp(de->d_name, ".") && strcmp(de->d_name, "..") &&
				prepare_path(pi, name, de->d_name, linkfirst))
			result = -1;
	closedir(dir);
	return result;
}

int prepare_single_path_info(struct path_info **pi, char *name,
		int linkfirst) {
	struct path_info **cpi;
	struct stat st;
	int result, link;
	link = 0;
	if (!path_is_safe(name)) {
		fprintf(stderr, "%s: %s: invalid path\n", progname, name);
		return -1;
	}
	if (lstat(name, &st)) {
		error(name);
		return -1;
	}
	for (cpi = pi; *cpi; cpi = &((*cpi)->next)) {
		if ((*cpi)->inode == st.st_ino && (*cpi)->device == st.st_dev) {
			if (linkfirst)
				(*cpi)->link = 1;
			else
				link = 1;
		}
	}
	*cpi = emalloc(sizeof(struct path_info), name);
	if (!*cpi)
		return -1;
	(*cpi)->next = NULL;
	(*cpi)->name = name;
	(*cpi)->device = st.st_dev;
	(*cpi)->inode = st.st_ino;
	(*cpi)->link = link;
	return !!S_ISDIR(st.st_mode);
}

int prepare_path(struct path_info **pi, char *parent, char *path,
		int linkfirst) {
	char *name;
	if (parent)
		name = construct_child_path(parent, path);
	else
		name = strdup(path);
	if (!name) {
		fprintf(stderr, "%s: %s/%s: out of memory\n", progname, parent, path);
		return -1;
	}
	switch (prepare_single_path_info(pi, name, linkfirst)) {
	case -1:
		free(name);
		return -1;
	case 0:
		return 0;
	case 1:
		return prepare_directory(pi, name, linkfirst);
	}
}

/* 0: success, -1: error */
int canonical_basename_dirname(char *s, char **dir, char **base) {
	char *t, *u, *v, *w, *x;
	int steps;
	t = strdup(s);
	if (!t)
		return -1;
	u = basename(t);
	v = strdup(s);
	if (!v) {
		free(t);
		return -1;
	}
	w = dirname(v);
	steps = 0;
	while ((!strcmp(u, ".") && strcmp(w, ".")) || !strcmp(u, "..")) {
		if (!strcmp(u, ".."))
			steps++;
		free(t);
		t = strdup(w);
		if (!t) {
			free(v);
			return -1;
		}
		u = basename(t);
		x = strdup(w);
		free(v);
		if (!x) {
			free(t);
			return -1;
		}
		v = x;
		w = dirname(v);
	}
	for (; steps; steps--) {
		if (!strcmp(u, ".")) {
			free(t);
			t = strdup("..");
			if (!t) {
				free(v);
				return -1;
			}
			u = t;
		} else if (!strcmp(u, "..")) {
			x = malloc(strlen(w) + 4);
			if (!x) {
				free(t);
				free(v);
				return -1;
			}
			sprintf(x, "%s/..", v);
			free(v);
			v = x;
			w = x;
		} else if (strcmp(w, "/")) {
			free(t);
			t = strdup(w);
			if (!t) {
				free(v);
				return -1;
			}
			u = basename(t);
			x = strdup(w);
			free(v);
			if (!x) {
				free(t);
				return -1;
			}
			v = x;
			w = dirname(v);
		}
	}
	*dir = strdup(w);
	free(v);
	if (!*dir) {
		free(t);
		return -1;
	}
	*base = strdup(u);
	free(t);
	if (!*base)
		return -1;
	return 0;
}

int paths_equal(char *s, char *t) {
	char *u, *v, *sb, *sd, *tb, *td;
	int r;
	u = strdup(s);
	if (!u)
		return -1;
	v = strdup(t);
	if (!v) {
		free(u);
		return -1;
	}
	while (strcmp(u, ".") && strcmp(v, ".") &&
			strcmp(u, "/") && strcmp(v, "/")) {
		canonical_basename_dirname(u, &sd, &sb);
		canonical_basename_dirname(v, &td, &tb);
		r = strcmp(sb, tb);
		free(u);
		free(v);
		free(sb);
		free(tb);
		if (r) {
			free(sd);
			free(td);
			return 0;
		}
		u = sd;
		v = td;
	}
	r = strcmp(u, v);
	free(u);
	free(v);
	return !r;
}

int path_is_untracked(char *path, struct stringlist *untracked) {
	struct stringlist *curr;
	char *s, *t;
	for (curr = untracked; curr; curr = curr->next) {
		s = strdup(path);
		if (!s) {
			fprintf(stderr, "%s: %s: out of memory\n", progname, path);
			exit(1);
		}
		while (1) {
			if (paths_equal(s, curr->s)) {
				free(s);
				return 1;
			}
			if (!strcmp(s, ".") || !strcmp(s, "/")) {
				free(s);
				break;
			}
			t = strdup(dirname(s));
			if (!t) {
				fprintf(stderr, "%s: %s: out of memory\n", progname, path);
				exit(1);
			}
			free(s);
			s = t;
		} 
	}
	return 0;
}

int output_path(struct stream *out, struct path_info *pi,
		struct path_info *opi, struct stringlist *untracked, uid_t uid,
		char *uname, gid_t gid, char *gname) {
	unsigned char pathbuf[101];
	unsigned char tgtbuf[101];
	unsigned char username[33];
	unsigned char groupname[33];
	char *target, *userptr, *groupptr;
	char blocktype;
	struct stat st;
	struct passwd *pw;
	struct group *gr;
	struct stream *in;
	struct path_info *cpi;
	off_t size;
	dev_t dev;
	size_t pathlen;
	ssize_t len;
	if (lstat(pi->name, &st)) {
		error(pi->name);
		return -1;
	}
	pathlen = strlen(pi->name);
	if (S_ISDIR(st.st_mode))
		pathlen++;
	if (pathlen > 100)
		output_long_link(out, pi->name);
	target = NULL;
	size = 0;
	if (S_ISDIR(st.st_mode))
		snprintf(pathbuf, 101, "%s/", pi->name);
	else
		snprintf(pathbuf, 101, "%s", pi->name);
	blocktype = '0';
	if (pi->link) {
		for (cpi = opi; cpi && cpi != pi; cpi = cpi->next) {
			if (cpi->inode == st.st_ino && !cpi->link) {
				blocktype = '1';
				snprintf(tgtbuf, 101, cpi->name);
				target = tgtbuf;
				break;
			}
		}
	}
	if (blocktype != '1') {
		if (S_ISDIR(st.st_mode)) {
			blocktype = '5';
		} else if (S_ISREG(st.st_mode)) {
				size = st.st_size;
				blocktype = '0';
		} else if (S_ISLNK(st.st_mode)) {
			blocktype = '2';
			if (st.st_size > 100) {
				fprintf(stderr, "%s: %s: link target too long\n", progname,
					pi->name);
				return -1;
			}
			len = readlink(pi->name, tgtbuf, 100);
			if (len < 0) {
				error(pi->name);
				return -1;
			}
			tgtbuf[len] = 0;
			target = tgtbuf;
		} else if (S_ISCHR(st.st_mode)) {
			blocktype = '3';
			dev = st.st_rdev;
		} else if (S_ISBLK(st.st_mode)) {
			blocktype = '4';
			dev = st.st_rdev;
		} else if (S_ISFIFO(st.st_mode)) {
			blocktype = '6';
		} else {
			fprintf(stderr, "%s: %s: invalid file type\n", progname, pi->name);
			return -1;
		}
	}
	if (uname) {
		userptr = uname;
	} else {
		pw = getpwuid(st.st_uid);
		if (pw) {
			snprintf(username, 33, "%s", pw->pw_name);
			userptr = username;
		} else {
			fprintf(stderr, "%s: %s: warning: invalid UID %d\n", progname,
				pi->name, st.st_uid);
			userptr = NULL;
		}
		uid = st.st_uid;
	}
	if (gname) {
		groupptr = gname;
	} else {
		gr = getgrgid(st.st_gid);
		if (gr) {
			snprintf(groupname, 33, "%s", gr->gr_name);
			groupptr = groupname;
		} else {
			fprintf(stderr, "%s: %s: warning: invalid GID %d\n", progname,
				pi->name, st.st_gid);
			groupptr = NULL;
		}
		gid = st.st_gid;
	}
	if (path_is_untracked(pathbuf, untracked)) {
		switch (blocktype) {
		case '0':
			blocktype = 'O';
			break;
		case '1':
			blocktype = 'l';
			break;
		case '2':
			blocktype = 'Z';
			break;
		case '3':
			blocktype = 'z';
			break;
		case '4':
			blocktype = 'y';
			break;
		case '5':
			blocktype = 's';
			break;
		case '6':
			blocktype = 'b';
			break;
		}
	}
	if (blocktype == '0' || blocktype == 'O') {
		in = file_stream_open(pi->name, "r");
		if (!in)
			return -1;
	}
	output_infoblock(out, pathbuf, st.st_mode & ~S_IFMT, uid, gid, size,
		st.st_mtime, blocktype, target, userptr, groupptr, dev);
	if (blocktype == '0' || blocktype == 'O') {
		output_file(in, out);
		in->close(in);
	}
	return 0;
}

void output_footer(struct stream *out) {
	fill_bytes(out, 1024);
}

int checksum_is_valid(char *buf, char *filename) {
	size_t i;
	int csum, scsum, dcsum;
	csum = checksum(buf, 0);
	scsum = checksum(buf, 1);
	dcsum = strtol(buf + 148, NULL, 8);
	if (dcsum != csum && dcsum != scsum) {
		fprintf(stderr, "%s: %s: bad checksum (expected %06o or %06o)\n",
			progname, filename, csum, scsum);
		return 0;
	}
	return 1;
}

/* 1: success, 0: no more headers */
int read_header(struct stream *in, char **filename, mode_t *mode, off_t *size,
		char *block_type, time_t *modtime, char **target, dev_t *dev) {
	char buf[512];
	char path[257];
	size_t len, fsize, tarsize;
	int devmajor, devminor;
	if (x_stream_read(in, buf, 512) < 512)
		return 0;
	if (!buf[0])
		return 0;
	if (!checksum_is_valid(buf, in->filename))
		exit(1);
	memcpy(path, buf + 345, 155);
	len = strlen(path);
	if (len && path[len - 1] != '/') {
		path[len] = '/';
		len++;
	}
	memcpy(path + len, buf, 100);
	path[len + 100] = 0;
	len = strlen(path);
	if (!strcmp(path, "././@LongLink")) {
		fsize = strtol(buf + 124, NULL, 8);
		tarsize = ((fsize + 511) & ~511);
		*filename = xmalloc(tarsize + 1, in->filename);
		if (x_stream_read(in, *filename, tarsize) < tarsize) {
			free(*filename);
			return 0;
		}
		(*filename)[tarsize] = 0;
		if (x_stream_read(in, buf, 512) < 512) {
			free(*filename);
			return 0;
		}
		if (!buf[0]) {
			free(*filename);
			return 0;
		}
		if (!checksum_is_valid(buf, in->filename))
			exit(1);
	} else if (!path_is_safe(path)) {
		fprintf(stderr, "%s: %s: invalid path\n", progname, path);
		exit(1);
	} else {
		*filename = strdup(path);
		if (!*filename) {
			fprintf(stderr, "%s: %s: out of memory\n", progname, in->filename);
			exit(1);
		}
	}
	*size = strtol(buf + 124, NULL, 8);
	if (mode)
		*mode = strtol(buf + 100, NULL, 8);
	if (block_type)
		*block_type = buf[156];
	if (modtime)
		*modtime = strtol(buf + 136, NULL, 8);
	if (target) {
		*target = xmalloc(101, in->filename);
		memcpy(*target, buf + 157, 100);
		(*target)[100] = 0;
	}
	if (dev) {
		devmajor = strtol(buf + 329, NULL, 8);
		devminor = strtol(buf + 337, NULL, 8);
		*dev = makedev(devmajor, devminor);
	}
	return 1;
}

void trim_dirname(char *name) {
	size_t len;
	len = strlen(name);
	if (len > 1 && name[len - 1] == '/') {
		name[len - 1] = 0;
		len--;
	}
	for (; len >= 2 && !strcmp(name + len - 2, "/."); len -= 2)
		name[len - 2] = 0;
}

int guarantee_dir(struct extractor *e, char *tgtdir, char *filename,
		mode_t mode, int track) {
	trim_dirname(filename);
	switch (e->mkdir(e, filename, mode, track)) {
	case -1:
		return -1;
	case 0:
		if (guarantee_parent_dir(e, tgtdir, filename, mode, track))
			return -1;
		return -(e->mkdir(e, filename, mode, track) != 1);
	case 1:
		return 0;
	}
}

int replace_open(char *filename, mode_t mode) {
	struct stat st;
	if (delete(filename, 0))
		return -1;
	else
		return open(filename, O_WRONLY|O_CREAT|O_NOFOLLOW|O_TRUNC, mode);
}

int ex_write(struct stream *in, char *filename, mode_t mode, time_t modtime,
		size_t size, int reopen, int flags, char *tgtdir) {
	struct stream *out;
	struct timeval tv[2];
	struct stat st;
	FILE *f;
	mode_t um;
	char *ffn;
	size_t tdlen, fnlen;
	int r, fd, oflags;
	r = 2;
	if (flags & EXTRACT_PERMISSION)
		um = umask(0);
	if (reopen) {
		fd = open(filename, O_WRONLY|O_NOFOLLOW);
	} else {
		oflags = O_WRONLY|O_CREAT|O_NOFOLLOW|O_TRUNC;
		if (flags & EXTRACT_KEEP)
			oflags |= O_EXCL;
		fd = open(filename, oflags, mode);
		if (fd < 0) {
			switch (errno) {
			case ENOENT:
				if (flags & EXTRACT_PERMISSION)
					umask(um);
				return 0;
			case ELOOP: case EISDIR: case EEXIST:
				if ((flags & EXTRACT_KEEP) && (flags & EXTRACT_NOISYKEEP))
					goto err0;
				if (!(flags & EXTRACT_KEEP))
					fd = replace_open(filename, mode);
				else
					return 1;
				break;
			case EACCES:
				if (flags & EXTRACT_KEEP)
					goto err0;
				else
					fd = replace_open(filename, mode);
				break;
			}
		}
	}
	if (fd < 0)
		goto err0;
	f = fdopen(fd, "w");
	if (!f) {
		close(fd);
		goto err0;
	}
	if (!tgtdir) {
		ffn = filename;
	} else {
		tdlen = strlen(tgtdir);
		fnlen = strlen(filename);
		if (tgtdir[tdlen - 1] == '/') {
			ffn = malloc(tdlen + fnlen + 1);
			if (!ffn) {
				target_emsg(tgtdir, filename, "out of memory");
				exit(1);
			}
			sprintf(ffn, "%s%s", tgtdir, filename);
		} else {
			ffn = malloc(tdlen + fnlen + 2);
			if (!ffn) {
				target_emsg(tgtdir, filename, "out of memory");
				exit(1);
			}
			sprintf(ffn, "%s/%s", tgtdir, filename);
		}
	}
	out = file_stream_new(f, ffn);
	if (!out) {
		fclose(f);
		if (ffn != filename)
			free(ffn);
		goto err;
	}
	if (copy_bytes(in, out, size) < size) {
		target_emsg(tgtdir, filename, "short read");
		exit(1);
	}
	if (fflush(f)) {
		target_error(tgtdir, filename);
		r = -1;
	}
	tv[0].tv_sec = time(NULL);
	tv[0].tv_usec = 0;
	tv[1].tv_sec = modtime;
	tv[1].tv_usec = 0;
	if (r >= 1 && (futimes(fd, tv) || (flags & EXTRACT_PERMISSION) &&
			(mode & 07000) && fchmod(fd, mode) && errno != EPERM)) {
		target_error(tgtdir, filename);
		r = -1;
	}
	if (flags & EXTRACT_PERMISSION)
		umask(um);
	out->close(out);
	if (ffn != filename)
		free(ffn);
	return r;
err0:
	target_error(tgtdir, filename);
err:
	if (flags & EXTRACT_PERMISSION)
		umask(um);
	x_stream_seek(in, size);
	return -1;
}

int ex_symlink(char *target, char *filename, int flags, char *tgtdir) {
	int r;
	r = 2;
	errno = 0;
	if (symlink(target, filename)) {
		switch (errno) {
		case ENOENT:
			return 0;
		case EEXIST:
			if ((flags & EXTRACT_KEEP) && (flags & EXTRACT_NOISYKEEP))
				goto err;
			errno = 0;
			if (!(flags & EXTRACT_KEEP)) {
				delete(filename, 0);
				symlink(target, filename);
			} else {
				r = 1;
			}
			break;
		}
	}
	if (errno)
		goto err;
	return r;
err:
	target_error(tgtdir, filename);
	return -1;
}

int ex_link(char *target, char *filename, int flags, char *tgtdir) {
	struct stat st;
	int r;
	r = 2;
	errno = 0;
	if (link(target, filename)) {
		switch (errno) {
		case ENOENT:
			if (stat(filename, &st))
				goto err;
			return 0;
		case EEXIST:
			if ((flags & EXTRACT_KEEP) && (flags & EXTRACT_NOISYKEEP))
				goto err;
			errno = 0;
			if (!(flags & EXTRACT_KEEP)) {
				delete(filename, 0);
				link(target, filename);
			} else {
				r = 1;
			}
			break;
		}
	}
	if (errno)
		goto err;
	return r;
err:
	target_error(tgtdir, filename);
	return -1;
}

int ex_mknod(char *filename, mode_t mode, dev_t dev, int flags, char *tgtdir) {
	mode_t um;
	int r;
	r = 2;
	errno = 0;
	if (flags & EXTRACT_PERMISSION)
		um = umask(0);
	if (mknod(filename, mode, dev)) {
		switch (errno) {
		case ENOENT:
			if (flags & EXTRACT_PERMISSION)
				umask(um);
			return 0;
		case EEXIST:
			if ((flags & EXTRACT_KEEP) && (flags & EXTRACT_NOISYKEEP))
				goto err;
			errno = 0;
			if (!(flags & EXTRACT_KEEP)) {
				delete(filename, 0);
				mknod(filename, mode, dev);
			} else {
				r = 1;
			}
			break;
		}
	}
	if (errno)
		goto err;
	if (flags & EXTRACT_PERMISSION)
		umask(um);
	return r;
err:
	if (flags & EXTRACT_PERMISSION)
		umask(um);
	target_error(tgtdir, filename);
	return -1;
}

int ex_mkdir(char *filename, mode_t mode, int flags, char *tgtdir) {
	struct stat st;
	int r;
	mode_t um;
	errno = 0;
	r = 2;
	if (flags & EXTRACT_PERMISSION)
		um = umask(0);
	if (mkdir(filename, mode)) {
		switch (errno) {
		case ENOENT: case ENOTDIR:
			if (flags & EXTRACT_PERMISSION)
				umask(um);
			return 0;
		case EEXIST:
			if ((flags & EXTRACT_KEEP) && (flags & EXTRACT_NOISYKEEP))
				goto err;
			errno = 0;
			if (!(flags & EXTRACT_KEEP)) {
				errno = 0;
				if (lstat(filename, &st))
					goto err;
				if (S_ISDIR(st.st_mode)) {
					chmod(filename, mode);
				} else {
					delete(filename, 0);
					mkdir(filename, mode);
				}
			} else {
				r = 1;
			}
			break;
		}
	}
	if (errno)
		goto err;
	if (flags & EXTRACT_PERMISSION)
		umask(um);
	return r;
err:
	if (flags & EXTRACT_PERMISSION)
		umask(um);
	target_error(tgtdir, filename);
	return -1;
}

int guarantee_parent_dir(struct extractor *e, char *tgtdir, char *filename,
		mode_t mode, int track) {
	char *s, *p, *t, *q;
	int r;
	s = strdup(filename);
	if (!s) {
		target_emsg(tgtdir, filename, "out of memory");
		return -1;
	}
	p = dirname(s);
	if (strcmp(p, "."))
		r = guarantee_dir(e, tgtdir, p, 0755, track);
	else
		r = 0;
	free(s);
	return r;
}

int extractor_write(struct extractor *e, struct stream *in, char *tgtdir,
		char *filename, mode_t mode, time_t modtime, size_t size, int track,
		int reopen) {
	int r;
	r = e->write(e, in, filename, mode, modtime, size, track, reopen);
	if (!r) {
		if (guarantee_parent_dir(e, tgtdir, filename, mode, track)) {
			x_stream_seek(in, size);
			return -1;
		}
		r = e->write(e, in, filename, mode, modtime, size, track, reopen);
	}
	if (r == 1)
		return 0;
	return -1;
}

int extractor_link(struct extractor *e, char *tgtdir, char *filename,
		char *target, int track) {
	int r;
	r = e->link(e, target, filename, track);
	if (!r) {
		if (guarantee_parent_dir(e, tgtdir, filename, 0777, track))
			return -1;
		r = e->link(e, target, filename, track);
	}
	if (r < 1)
		return -1;
	return 0;
}

int extractor_symlink(struct extractor *e, char *tgtdir, char *filename,
		char *target, int track) {
	int r;
	r = e->symlink(e, target, filename, track);
	if (!r) {
		if (guarantee_parent_dir(e, tgtdir, filename, 0777, track))
			return -1;
		r = e->symlink(e, target, filename, track);
	}
	if (r < 1)
		return -1;
	return 0;
}

int extractor_mknod(struct extractor *e, char *tgtdir, char *filename,
		dev_t dev, mode_t mode, int track) {
	int r;
	r = e->mknod(e, filename, mode, dev, track);
	if (!r) {
		if (guarantee_parent_dir(e, tgtdir, filename, mode, track))
			return -1;
		r = e->mknod(e, filename, mode, dev, track);
	}
	if (r < 1)
		return -1;
	return 0;
}

int extract_file(struct extractor *e, char *tgtdir, struct stream *in,
		int trackable) {
	char *filename, *target, *msg;
	mode_t mode;
	off_t size;
	time_t modtime;
	dev_t dev;
	int result, track;
	char block_type;
	result = 1;
	if (!read_header(in, &filename, &mode, &size, &block_type, &modtime,
			&target, &dev))
		return 0;
	track = !block_type || (block_type >= '0' && block_type <= '9') ||
		block_type == 'g' || block_type == 'x';
	if (!track && !trackable)
		goto uft;
	switch (block_type) {
	case '0': case 0: case 'O':
		if (extractor_write(e, in, tgtdir, filename, mode, modtime, size,
				track, 0))
			result = -1;
		x_stream_seek(in, (512 - size % 512) % 512);
		break;
	case '1': case 'l':
		if (extractor_link(e, tgtdir, filename, target, track))
			result = -1;
		break;
	case '2': case 'Z':
		if (extractor_symlink(e, tgtdir, filename, target, track))
			result = -1;
		break;
	case '3': case 'z':
		if (extractor_mknod(e, tgtdir, filename, dev, S_IFCHR | mode, track))
			result = -1;
		break;
	case '4': case 'y':
		if (extractor_mknod(e, tgtdir, filename, dev, S_IFBLK | mode, track))
			result = -1;
		break;
	case '5': case 's':
		if (guarantee_dir(e, tgtdir, filename, mode, track))
			result = -1;
		break;
	case '6': case 'b':
		if (extractor_mknod(e, tgtdir, filename, dev, S_IFIFO | mode, track))
			result = -1;
		break;
	case 'g':
	case 'x':
		fprintf(stderr, "%s: warning: PAX header ignored\n", progname);
		x_stream_seek(in, size + (512 - size % 512) % 512);
		break;
	default:
		goto uft;
	}
	goto cleanup;
uft:
	msg = malloc(22);
	if (msg) {
		sprintf(msg, "unknown file type '%c'", block_type);
		target_emsg(tgtdir, filename, msg);
		free(msg);
	} else {
		target_emsg(tgtdir, filename, "unknown file type");
	}
	result = -1;
	x_stream_seek(in, size + (512 - size % 512) % 512);
cleanup:
	free(filename);
	free(target);
	return result;
}

int output_escaped_string(FILE *f, char *s) {
	for (; *s; s++) {
		switch (*s) {
		case '\\':
			if (fputc('\\', f) < 0 || fputc('\\', f) < 0)
				return -1;
			break;
		case '\n':
			if (fputc('\\', f) < 0 || fputc('n', f) < 0)
				return -1;
			break;
		default:
			if (fputc(*s, f) < 0)
				return -1;
			break;
		}
	}
	return 0;
}

int write_kvp(FILE *f, char *key, char *value) {
	if (output_escaped_string(f, key))
		return -1;
	if (fputc(':', f) < 0 || fputc(' ', f) < 0)
		return -1;
	if (output_escaped_string(f, value))
		return -1;
	if (fputc('\n', f) < 0)
		return -1;
	return 0;
}

int read_kvp(FILE *f, char *filename, char **key, char **value) {
	size_t key_size, value_size;
	int mode, ch, pos;
	key_size = 16;
	*key = xmalloc(key_size, filename);
	value_size = 16;
	*value = xmalloc(value_size, filename);
	mode = 0;
	pos = 0;
	for (ch = fgetc(f); ch >= 0 && ch != 10; ch = fgetc(f)) {
		switch (mode) {
		case 0:
			switch (ch) {
			case ':':
				(*key)[pos] = 0;
				mode = 1;
				break;
			case '\\':
				mode = 3;
				break;
			default:
				(*key)[pos] = ch;
				pos++;
			}
			if (pos == key_size) {
				key_size *= 2;
				*key = xrealloc(*key, key_size, filename);
			}
			break;
		case 1:
			switch (ch) {
			case ' ': case '\t':
				break;
			case '\\':
				pos = 0;
				mode = 4;
				break;
			default:
				(*value)[0] = ch;
				pos = 1;
				mode = 2;
			}
			break;
		case 2:
			if (ch == '\\') {
				mode = 4;
			} else {
				(*value)[pos] = ch;
				pos++;
				if (pos == value_size) {
					value_size *= 2;
					*value = xrealloc(*value, value_size, filename);
				}
			}
			break;
		case 3:
			if (ch == 'n')
				(*key)[pos] = '\n';
			else
				(*key)[pos] = ch;
			pos++;
			if (pos == key_size) {
				key_size *= 2;
				*key = xrealloc(*key, key_size, filename);
			}
			mode = 0;
			break;
		case 4:
			if (ch == 'n')
				(*value)[pos] = '\n';
			else
				(*value)[pos] = ch;
			pos++;
			if (pos == value_size) {
				value_size *= 2;
				*value = xrealloc(*value, value_size, filename);
			}
			mode = 2;
			break;
		}
	}
	if (mode == 0 || mode == 3)
		(*key)[pos] = 0;
	else if (mode == 2 || mode == 4)
		(*value)[pos] = 0;
	return mode == 2 || mode == 4;
}

struct tar_extractor {
	struct extractor e;
	int flags;
};

int tar_extractor_write(struct extractor *e, struct stream *in, char *filename,
		mode_t mode, time_t modtime, size_t size, int track, int reopen) {
	struct tar_extractor *te;
	int r;
	te = (struct tar_extractor *)e;
	r = ex_write(in, filename, mode, modtime, size, reopen, te->flags, NULL);
	if (r < 1) {
		return r;
	} else {
		if (r == 1)
			x_stream_seek(in, size);
		return 1;
	}
}

int list_extractor_write(struct extractor *e, struct stream *in,
		char *filename, mode_t mode, time_t modtime, size_t size, int track,
		int reopen) {
	x_stream_seek(in, size);
	printf("%s\n", filename);
	return -1;
}

int tar_extractor_symlink(struct extractor *e, char *target, char *filename,
		int track) {
	struct tar_extractor *te;
	int r;
	te = (struct tar_extractor *)e;
	r = ex_symlink(target, filename, te->flags, NULL);
	if (r < 1)
		return r;
	else
		return 1;
}

int list_extractor_symlink(struct extractor *e, char *target, char *filename,
		int track) {
	printf("%s\n", filename);
	return 1;
}

int tar_extractor_link(struct extractor *e, char *target, char *filename,
		int track) {
	struct tar_extractor *te;
	int r;
	te = (struct tar_extractor *)e;
	r = ex_link(target, filename, te->flags, NULL);
	if (r < 1)
		return r;
	else
		return 1;
}

int list_extractor_link(struct extractor *e, char *target, char *filename,
		int track) {
	printf("%s\n", filename);
	return 1;
}

int tar_extractor_mknod(struct extractor *e, char *filename, mode_t mode,
		dev_t dev, int track) {
	struct tar_extractor *te;
	int r;
	te = (struct tar_extractor *)e;
	r = ex_mknod(filename, mode, dev, te->flags, NULL);
	if (r < 1)
		return r;
	else
		return 1;
}

int list_extractor_mknod(struct extractor *e, char *filename, mode_t mode,
		dev_t dev, int track) {
	printf("%s\n", filename);
	return 1;
}

int tar_extractor_mkdir(struct extractor *e, char *filename, mode_t mode,
		int track) {
	struct tar_extractor *te;
	int r;
	te = (struct tar_extractor *)e;
	r = ex_mkdir(filename, mode, te->flags, NULL);
	if (r < 1)
		return r;
	else
		return 1;
}

int list_extractor_mkdir(struct extractor *e, char *filename, mode_t mode,
		int track) {
	printf("%s\n", filename);
	return 1;
}

struct extractor *new_tar_extractor(int flags) {
	struct tar_extractor *te;
	te = malloc(sizeof(struct tar_extractor));
	if (!te) {
		fprintf(stderr, "%s: out of memory\n", progname);
		return NULL;
	}
	te->flags = flags;
	te->e.write = tar_extractor_write;
	te->e.symlink = tar_extractor_symlink;
	te->e.link = tar_extractor_link;
	te->e.mknod = tar_extractor_mknod;
	te->e.mkdir = tar_extractor_mkdir;
	return (struct extractor *)te;
}

struct extractor *new_list_extractor() {
	struct extractor *e;
	e = malloc(sizeof(struct extractor));
	if (!e) {
		fprintf(stderr, "%s: out of memory\n");
		return NULL;
	}
	e->write = list_extractor_write;
	e->symlink = list_extractor_symlink;
	e->link = list_extractor_link;
	e->mknod = list_extractor_mknod;
	e->mkdir = list_extractor_mkdir;
	return e;
}

void target_emsg(char *target, char *filename, char *emsg) {
	if (!target)
		fprintf(stderr, "%s: %s: %s\n", progname, filename, emsg);
	else if (target[strlen(target) - 1] == '/')
		fprintf(stderr, "%s: %s%s: %s\n", progname, target, filename, emsg);
	else
		fprintf(stderr, "%s: %s/%s: %s\n", progname, target, filename, emsg);
}

void target_error(char *target, char *filename) {
	target_emsg(target, filename, strerror(errno));
}

void target_die(char *target, char *filename) {
	target_error(target, filename);
	exit(1);
}
